# !/usr/bin/env python

import matplotlib.pyplot as plt
import seaborn as sns

from omniplate.platereader import platereader

__version__ = "2.11.1"

plt.rcParams["figure.max_open_warning"] = 0
sns.set()
